import os
import sys
from ..utils import logger


class PowerFactoryConn:

    def __init__(self, 
                 path, 
                 username,
                 password,
                 logger=logger):

        self.path = path
        self.username = username
        self.password = password
        self.logger = logger
        
        self.check_path()
        self.check_credentials()

    def check_path(self):
        if not os.path.isdir(self.path):
            raise Exception("Invalid power factory python directory")

    def check_valid_conn(self, app):
        if app is None:
            raise Exception("Power Factory Connection failed.. please check your credentials")
        else:
            self.logger.info("Power Factory connected sucessfully")
    
    def check_credentials(self):
        if not self.username:
            raise Exception(":: Username must be provided")
        if self.password is None:
            self.logger.warning("Password not provided")

    def connect(self):

        try:
            sys.path.append(f"{self.path}")
            
            import powerfactory
            logger.info("Logging to Power Factory.. please wait")        
            app = powerfactory.GetApplication(self.username,
                                              self.password)
            self.check_valid_conn(app)
            self.logger.info("Connected with user -{}-".format(self.username))
         
        except Exception as e:
            print(e)
        
        return app


import os
import sys
import pandas as pd
import numpy as np


class BusParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        bus_dict = {}
        buses = self.app.GetCalcRelevantObjects('*.ElmTerm')
        for bus in buses:
            params = {
                "energized": bus.ciEnergized,
                "usage": bus.iUsage,
                "uknom": bus.uknom,
                "vmax": bus.vmax,
                "vmin": bus.vmin,
            }
            bus_dict[bus.loc_name] = params        
        return pd.DataFrame.from_dict(bus_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

class LineParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        line_dict = {}
        lineselem = self.app.GetCalcRelevantObjects('*.ElmLne')
        linestype = self.app.GetCalcRelevantObjects('*.TypLne')
        for lineelem, linetype in zip(lineselem, linestype):

            try:
                node_bus1 = lineelem.bus1.cBusBar.loc_name
            except:
                node_bus1 = np.nan
            try:
                node_bus2 = lineelem.bus2.cBusBar.loc_name
            except:
                node_bus2 = np.nan

            params = {
                "out_service": lineelem.IsOutOfService(),
                "node_bus1": node_bus1,
                "node_bus2": node_bus2,
                "unom": lineelem.Unom,
                "length": lineelem.dline,
                "rated_cur": linetype.sline,
                "type": linetype.cohl_,
                "freq": linetype.frnom,
                "r12": linetype.rline,
                "x12": linetype.xline,
                "r0": linetype.rline0,
                "x0": linetype.xline0,
                "b12": linetype.bline,
                "b0": linetype.bline0,
                # "z1": line.Z1,
                # "r0": line.R0,
                # "r1": line.R1,
                # "x0": line.X0,
                # "x1": line.X1,
                # "b0": line.B0,
                # "b1": line.B1,
                # "c0": line.C0,
                # "c1": line.C1,
                # "g0": line.G0,
                # "g1": line.G1,
            }
            line_dict[lineelem.loc_name] = params    

        return pd.DataFrame.from_dict(line_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

class SynGenParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        gen_dict = {}
        gens_elem = self.app.GetCalcRelevantObjects('*.ElmSym')
        gens_type = self.app.GetCalcRelevantObjects('*.TypSym')
        for genelem, gentype in zip(gens_elem, gens_type):

            try: 
                node = genelem.bus1.cBusBar.loc_name
            except:
                node = np.nan

            params = {
                "out_service": genelem.IsOutOfService(),
                "node": node,
                "unom": genelem.GetUnom(),
                "pmax": genelem.P_max,
                "pmin": genelem.P_min,
                "snom": gentype.sgn,
                "unom": gentype.ugn,
                "qmax": gentype.Q_max,
                "qmin": gentype.Q_min,
                "pfactor": gentype.cosn,
                "r0_pu": gentype.r0sy,
                "x0_pu": gentype.x0sy,
                "r2_pu": gentype.r2sy,
                "x2_pu": gentype.x2sy,
                "xd": gentype.xd,
                "xq": gentype.xq,
                "rotor_type": gentype.iturbo,
            }
            gen_dict[genelem.loc_name] = params 

        return pd.DataFrame.from_dict(gen_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

class Trafo3wParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        trafo_dict = {}
        trafos_elem = self.app.GetCalcRelevantObjects('*.ElmTr3')
        trafos_type = self.app.GetCalcRelevantObjects('*.TypTr3')
        for trafoelm, trafotype in zip(trafos_elem, trafos_type):
            
            try:
                node_hv = trafoelm.bushv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_mv = trafoelm.busmv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_lv = trafoelm.buslv.cBusBar.loc_name
            except:
                node_hv = np.nan

            params = {
                "out_service": trafoelm.IsOutOfService(),
                "node_hv": node_hv,
                "node_mv": node_mv,
                "node_lv": node_lv,
                "v_hv": trafotype.utrn3_h,
                "v_mv": trafotype.utrn3_m,
                "v_lv": trafotype.utrn3_l,
                "s_hv": trafotype.strn3_h,
                "s_mv": trafotype.strn3_m,
                "s_lv": trafotype.strn3_l,
                "r_hv_pu": trafotype.r1pu_h,
                "x_hv_pu": trafotype.x1pu_h,
                "r_mv_pu": trafotype.r1pu_m,
                "x_mv_pu": trafotype.x1pu_m,
                "r_lv_pu": trafotype.r1pu_l,
                "x_lv_pu": trafotype.x1pu_l,
            }
            trafo_dict[trafoelm.loc_name] = params  

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

class Trafo2wParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        trafo_dict = {}
        trafos_elem = self.app.GetCalcRelevantObjects('*.ElmTr2')
        trafos_type = self.app.GetCalcRelevantObjects('*.TypTr2')
        for trafoelm, trafotype in zip(trafos_elem, trafos_type):

            try:
                node_hv = trafoelm.bushv.cBusBar.loc_name
            except: 
                node_hv = np.nan
            try:
                node_lv = trafoelm.buslv.cBusBar.loc_name
            except:
                node_lv = np.nan

            params = {
                "out_service": trafoelm.IsOutOfService(),
                "node_hv": node_hv,
                "node_lv": node_lv,
                "snom": trafotype.strn,
                "freq": trafotype.frnom,
                "v_hv": trafotype.utrn_h,
                "v_lv": trafotype.utrn_l,
                "z1_perc": trafotype.uktr,
                "z0_perc": trafotype.uk0tr,
            }
            trafo_dict[trafoelm.loc_name] = params      

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

class ShuntParam():

    def __init__(self, app):
        self.app = app

    def build_df(self):
        
        shunt_dict = {}
        shunts_elem = self.app.GetCalcRelevantObjects('*.ElmShnt')
        for shunt in shunts_elem:

            try:
                node = shunt.bus1.cBusBar.loc_name
            except: 
                node = np.nan

            params = {
                "out_service": shunt.IsOutOfService(),
                "node": node,
                "unom": shunt.ushnm,
                "qnom": shunt.qcapn,
                "type": shunt.shtype, 
            }
            shunt_dict[shunt.loc_name] = params 

        return pd.DataFrame.from_dict(shunt_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})


import os
import sys
from ..utils import logger


class ShcEvents:

    def __init__(self, app):
        self.app = app
        self.clearance_suffix = "_clearfault"

    def __valid_event(self, event):
        cond = True if isinstance(event, dict) else False
        if not cond:
            raise Exception(f"Event:=> {event} passed with wrong format")

    def create(self, event):

        # Validate event schema
        self.__valid_event(event)
        is_event_created = False

        for k, payload in event.items():
            
            # Get event name
            event_name = k

            # Check target element 
            target_elem, elem_type = payload["target_elem"][0], \
                                     payload["target_elem"][1]
            types = {
                "line": ".ElmLne",
                "gen": ".ElmSym",
            }
            full_target_elem = f"{target_elem}{types[elem_type]}"
            target = self.app.GetCalcRelevantObjects(full_target_elem)[0]
            if not target:
                raise Exception("Target element neither found it nor provided it")

            # Enable RMS sim (only lines)
            if elem_type == "line":
                target.ishclne = 1
                try:
                    fault_perc = payload["fault_percen"]
                    target.fshcloc = fault_perc
                except:
                    logger.warn("Fault percentage not detected in your event.. set up as 50%")
                    fault_perc = 50
                    target.fshcloc = fault_perc
                logger.warn(f"Fault will hit the line at {round(target.cshcloc, 1)} [km] or {fault_perc}% from S/E {target.bus1.cBusBar.loc_name}")

            # Check fault type
            try:
                fault = {
                    "3F": 0,
                    "2F": 1,
                    "1FG": 2,
                }
                fault_name = payload["fault_type"]
                fault_num = fault[fault_name]
            except: 
                logger.warn("Fault type not attached at your event.. setting it as 3 Phase Fault")
                fault_num = 0
                fault_name = "3F"

            # Check start time
            try:
                fault_time = payload["fault_duration"]
            except:
                fault_time = 0.3
                logger.warn(f"Fault duration not attached at your event.. setting it as {fault_time}")

            # Clearance fault
            try:
                clearance_time = payload["clearance_time"]
                clearance = True
            except:
                logger.warn("Clearance not attached at your event")
                clearance = False
        
        # Creating event in power factory
        event_folder = self.app.GetFromStudyCase("IntEvt")
        event_folder.CreateObject("EvtShc", event_name)
        my_event = event_folder.GetContents(event_name + ".EvtShc")[0]
        if my_event:
            logger.info("Event Created with the following information:")
            logger.info(f"\tname: -{event_name}-")
            logger.info(f"\ttarget_elem: -{target_elem}- and fault: {fault_name}")
            logger.info(f"\tclearance: {clearance} with time: {clearance_time}")
        
        # Add event properties
        my_event.time = fault_time
        my_event.p_target = target
        my_event.i_shc = fault_num

        if clearance:
            clearance_name = f"{event_name}{self.clearance_suffix}"
            event_folder.CreateObject("EvtShc", clearance_name)
            my_clearance_event = event_folder.GetContents(clearance_name + ".EvtShc")[0]
            my_clearance_event.time = fault_time + clearance_time
            my_clearance_event.p_target = target
            my_clearance_event.i_sch = 4
        
        # Validate even creation
        is_event_created = True
        
        return is_event_created

    def delete(self, event):

        # Validate event schema
        self.__valid_event(event)
        is_event_created = False

        for k, v in event.items():
            
            # Get event name
            event_name = k

            evname = event_name + ".EvtShc"
            clearance_evname = event_name + self.clearance_suffix + ".EvtShc"
            event_folder = self.app.GetFromStudyCase("IntEvt")
            my_event = event_folder.GetContents(evname)[0]
            my_clearance_event = event_folder.GetContents(clearance_evname)[0]

            try:
                my_event.Delete()
                my_clearance_event.Delete()
            except:
                logger.warn("Events not deleted")

        return None

    def pause(self, event):

         # Validate event schema
        self.__valid_event(event)
        is_event_created = False

        for k, v in event.items():
            
            # Get event name
            event_name = k

            evname = event_name + ".EvtShc"
            clearance_evname = event_name + self.clearance_suffix + ".EvtShc"
            event_folder = self.app.GetFromStudyCase("IntEvt")
            my_event = event_folder.GetContents(evname)[0]
            my_clearance_event = event_folder.GetContents(clearance_evname)[0]

            try:
                my_event.outserv = 1
                my_clearance_event.outserv = 1
            except:
                logger.warn("Events not paused")

    def clear_all(self):

        event_folder = self.app.GetFromStudyCase("IntEvt")
        events = event_folder.GetContents()
        if len(events) > 0:
            logger.warn(f"Cleaning all your events in your SCase.. tot num of events detected:=> {len(events)}")
        else:
            logger.info("Event folder seems to be clean.. nothing to delete")

        for event in events:
            try:
                event.Delete()
            except:
                logger.warn("Event not deleted.. there will be some remaining events")
        
        return None


import os
import sys
import re
import pandas as pd
import numpy as np

from ..utils import logger


class CleanDf:

    def __init__(self, df):
        self.df = df
    
    def clean_name(self, list_):
        return [re.sub('\.[0-9]', '', x).strip() for x in list_]

    def clean_metric(self, list_):
        pattn = r"([a-z]{1}):([a-zA-Z]+)(.)"
        return [re.match(pattn, x).group(2).lower() for x in list_]

    def clean_headers(self):

        # Transform elements name into level 0
        elems_name = self.df.columns
        col_level_0 = self.clean_name(elems_name)
        col_level_0[0] = "t"

        # Transform metric o var name into level 1
        vars_name = self.df.loc[0]
        col_level_1 = self.clean_metric(vars_name)

        cols = [(x, y)for x, y in zip(col_level_1, col_level_0)]
        self.df.columns = pd.MultiIndex.from_tuples(cols)
        self.df.drop(self.df.index[0], inplace=True)

        return self.df

class ResDynSim:

    def __init__(self, app):
        
        self.app = app
        self.res = None
        self.elmResfile = "TmpSimRel.ElmRes"
        self.__path = "tmp"
        self.__df = None

    def _prepare_results(self):

        # Create ElmRes
        logger.info(f"Creating a temporary {self.elmResfile} file in your StudyCase")
        self.res = self.app.GetFromStudyCase(f'{self.elmResfile}')

        if self.res:
            # logger.info("Cleaning ElmRes")
            self.res.Clear()
            self.res.SetAsDefault()
        else:
            raise Exception(f"{self.elmResfile} creation failed")

        # Add variable of interest
        logger.info(f"Adding variables to .ElmRes -> It might take few mins")
        self.buses = self.app.GetCalcRelevantObjects("*.ElmTerm")
        self.gens = self.app.GetCalcRelevantObjects("*.ElmSym")
        # self.lines = self.app.GetCalcRelevantObjects("*.ElmLne")

        # Adding variables
        self.__bus_vars = ["m:u", "m:phiu", "m:fehz"]
        
        _ = [self.res.AddVars(bus, "m:u",
                                   "m:phiu",
                                   "m:fehz",
                                   ) for bus in self.buses]

        self.__gen_vars = ["c:firel", "s:P1", "s:Q1", "s:ut", "s:xspeed"]

        _ = [self.res.AddVars(gen, "c:firel", 
                                   "s:P1", 
                                   "s:Q1",
                                   "s:ut", 
                                   "s:xspeed",
                                   ) for gen in self.gens]

        return None

    def _get_results(self):

        # Load ElmRes File
        self.app.ResLoadData(self.res)
        self.rows_num = self.app.ResGetValueCount(self.res, 0)

        is_exp_ok = False
        try:
            # Temporary fix (export as csv)
            comres = self.app.GetFromStudyCase('ComRes')
            comres.iopt_csel = 1
            comres.iopt_tsel = 0
            comres.iopt_locn = 1
            comres.ciopt_head = 1
            comres.iopt_exp = 6
            comres.pResult = self.res

            for case in ["gen", "bus"]:
                
                f_name = f"{self.__path}_{case}.csv"
                comres.f_name = f_name

                if case == "gen":
                    elements = [(self.res, gen, var) for gen in self.gens for var in self.__gen_vars]
                    elements.insert(0, (self.res, self.res, "b:tnow"))
                    comres.resultobj = [el[0] for el in elements]
                    comres.element = [el[1] for el in elements]
                    comres.variable = [el[2] for el in elements]
                    comres.Execute()

                    try:
                        self.__df_gen = pd.read_csv(f_name, engine='python')
                    except Exception as e:
                        print(e)

                if case == "bus":
                    elements = [(self.res, bus, var) for bus in self.buses for var in self.__bus_vars]
                    elements.insert(0, (self.res, self.res, "b:tnow"))
                    comres.resultobj = [el[0] for el in elements]
                    comres.element = [el[1] for el in elements]
                    comres.variable = [el[2] for el in elements]
                    comres.Execute()
                    
                    try:
                        self.__df_bus = pd.read_csv(f_name, engine='python')
                    except Exception as e:
                        print(e)

                # Remove tmp csv file
                os.remove(f_name)

            is_exp_ok = True

        except Exception as e:
            print(e)

        return is_exp_ok

    def _clean_gen(self):
        df_inst = CleanDf(self.__df_gen)
        return df_inst.clean_headers()

    def _clean_bus(self):
        df_inst = CleanDf(self.__df_bus) 
        return df_inst.clean_headers()

    @property
    def gen(self):
        return self._clean_gen()

    @property
    def bus(self):
        return self._clean_bus()


import sys
import os
import re
import time

from .app.conn import PowerFactoryConn
from .app.params import BusParam, \
                        LineParam, \
                        SynGenParam, \
                        Trafo3wParam, \
                        Trafo2wParam, \
                        ShuntParam
from .app.res import ResLdf
from .app.res import ResDynSim
from .app.events import ShcEvents
from .app.utils import logger


class Network:

    def __init__(self, path, username=None, password=None):
        self.app = PowerFactoryConn(path, username, password).connect()

    def __validated_study_case_activation(self):
        if self.__studycase_activated:
            logger.info("Okay.. study case is activated")
        else:
            # raise Exception("Please first select a study case")
            logger.error("Please first select a study case")
        return None

    def show_power_factory(self):
        self.app.Show()

    def activate_project(self, project):

        """Activate a project given a input name

        Returns:
            [bool]: succeed or not
        """
        
        try:
            logger.info("Trying to locate your project.. please wait")
            activate_proj = self.app.ActivateProject(project)
            if activate_proj == 1:
                self.__proj_activated = False
                logger.error("Project not found")
                # sys.exit()
            else:
                logger.info(f"Project:=> -{project}- activated")
                self.__proj_activated = True
        except:
            raise Exception ("Error when selecting project case")
        
        return self.__proj_activated

    def activate_study_case(self, study_case):

        """Activate a study case given a input name

        Returns:
            [bool]: succeed or not
        """

        try:
            logger.info("Getting your stude case.. please wait")
            study_case_folder = self.app.GetProjectFolder("study")
            all_study_cases = study_case_folder.GetContents(f"{study_case}.IntCase")
            if not all_study_cases:
                self.__studycase_activated = False
                logger.error("Study case not found")
            else:
                study_case_pf = all_study_cases[0]
                study_case_pf.Activate
                logger.info(f"Selected study case:=> -{study_case}-")
                self.__studycase_activated = True
        except:
            raise Exception ("Error when selecting study case")

        return self.__studycase_activated

    def get_bus_params(self):

        """
        Returns:
            [dataframe]: bus grid values
        """
        self.__validated_study_case_activation()
        return BusParam(self.app).build_df()
       
    def get_line_params(self):

        """
        Returns:
            [dataframe]: line grid values
        """
        self.__validated_study_case_activation()
        return LineParam(self.app).build_df()

    def get_gen_params(self):

        """
        Returns:
            [dataframe]: gen grid values
        """
        self.__validated_study_case_activation()
        return SynGenParam(self.app).build_df()

    def get_3wtrafo_params(self):

        """
        Returns:
            [dataframe]: 3 wind trafo grid values
        """
        self.__validated_study_case_activation()
        return Trafo3wParam(self.app).build_df()

    def get_2wtrafo_params(self):

        """
        Returns:
            [dataframe]: 2 wind trafo grid values
        """
        self.__validated_study_case_activation()
        return Trafo2wParam(self.app).build_df()

    def get_shunt_params(self):
        
        """
        Returns:
            [dataframe]: shunt grid values
        """
        self.__validated_study_case_activation()
        return ShuntParam(self.app).build_df()
    
    def get_all_projects(self):
        raise Exception("Not implemented yet")

    def search_element_name(self, name, elem_type=None):

        """ Find all possible elem name

        Args:
            name [str]:
                element name to find
            elem_type [str]:
                options: "line", "load", "gen", "3wtrafo", "2wtrafo", "shunt"

        Returns:
            [list]: all possible matches
        """
        
        elem_types = ["line", "load", "gen", "3wtrafo", "2wtrafo", "shunt"]
        if elem_type not in elem_types:
            raise Exception(f"elem_type argument not valid.. possible values {elem_types}")

        if elem_type == "line":
            logger.info("Searching for elements in lines")
            arg = f"{name}.ElmLne"

        if elem_type == "load":
            logger.info("Searching for elements in loads")
            arg = f"{name}.ElmLod"

        if elem_type == "gen":
            logger.info("Searching for elements in sync generators")
            arg = f"{name}.ElmSym"

        if elem_type == "3wtrafo":
            logger.info("Searching for elements in 3 winds transformers")
            arg = f"{name}.ElmTr3"

        if elem_type == "2wtrafo":
            logger.info("Searching for elements in 2 winds transformers")
            arg = f"{name}.ElmTr2"

        if elem_type == "shunt":
            logger.info("Searching for elements in shunts")
            arg = f"{name}.ElmShnt"

        arg_, type_ = arg.split(".") 
        arg_lower = "*" + arg_.lower() + "*" + "." + type_
        arg_upper = "*" + arg_.upper() + "*" + "." + type_
        arg_title = "*" + arg_.title() + "*" + "." + type_
        args = [arg_lower, arg_upper, arg_title]
        elems_tot = []
        for arg in args:
            elems = self.app.GetCalcRelevantObjects(arg)
            elems_found = [elem.loc_name for elem in elems]
            elems_tot.extend(elems_found)
        logger.info(f"Total num of elements found:=> {len(elems_tot)}")

        return elems_tot

    def run_load_flow(self, 
                      ldf_mode="balanced",
                      nraph_mode="classical",
                      plimits=False,
                      qlimits=False,
                      show=False,
                      ):

        """Run Load Flow simulation

         Optional Args:
            ldf_mode [str]: -balanced- default
                options: balanced, unbalanced, dc
            nraph_mode [str]: -classical- default
                options: classical, current
            plimits [bool]: -False- default
                consider power limits or not
            qlimits [bool]: -False- default
                consider reactive limits or not
            show [bool]: -False- default
                display power factory or not

        Returns:
            [dataframe]: class than contains dataframes organized by elem types
            Example:
                res.gen
                res.gen_pu
                res.bus
                res.bus_pu
                res.line
                res.line_pu
                res.trafo3w
                res.trafo3w_pu
                res.trafo2w
                res.trafo2w_pu
                res.shunt
        """

        # Get load flow mode
        modes = {
            "balanced": 0,
            "unbalanced": 1,
            "dc": 2,
        }
        nraphson = {
            "classical": 1,
            "current": 0,
        }
        plim = 1 if plimits else 0
        qlim = 1 if qlimits else 0

        is_ldf_ok = False
        if self.__studycase_activated == True:

            logger.info("Running power flow.. please wait")
            if show:
                self.show_power_factory()

            # Run power flow with options
            ldf = self.app.GetFromStudyCase('ComLdf')
            ldf.iopt_net = modes[ldf_mode]
            ldf.i_power = nraphson[nraph_mode]
            ldf.iopt_plim = plim
            ldf.iopt_lim = qlim

            resp = ldf.Execute()

            if resp == 0:
                logger.info("Load flow - OK - ")
                is_ldf_ok = True
            if resp == 1:
                logger.error("Load flow failed due to divergene of inner loops")
            if resp == 2:
                logger.error("Load flow failed due to divergene of outer loops")

        else:
            logger.warn("Load flow cannot executed .. first activate a study case")
        
        return ResLdf(self.app, is_ldf_ok)

    def run_dynamic_sim(self,
                        sim_type="rms",
                        start_time=0,
                        end_time=5,
                        step=0.01,
                        events=[],
                        show=False,
                        del_events=False,
                       ):
        
        """Run dynamic simulation (only rms for now)

        Optional Args:
            sim_type [str]: -rms- default
                rms/ins
            start_time [seg]: -0- default
                time when simu starts
            end_time [seg]:  -5- default
                time when simu ends
            step [float]: -0.01- default
                step value for simulation
            show [bool]: -False- default
                display power factory or not
            del_events [bool]: -False- default 
                del your events after all computation if True / paused if False
            events [list]: 
                list of your events 
                Example:
                events = [
                    {"my_event": {"target_elem": ("L_DURA_ESCL_2_1", "line"),
                                  "fault_duration": 0.05,
                                  "clearance_time": 0.01,
                                  "fault_type": "1FG"}
                                 },
                ]

        Returns:
            [dataframe] -- class that constains gen and bus results
            Example:
                res.gen / res.bus
        """

        is_sim_ok = False
        if self.__studycase_activated == True:

            logger.info(f"Initializing {sim_type} dynamic simulation.. please wait")
            if show:
                self.show_power_factory()
            
            # Parsing events
            num = 0
            if not events:
                logger.warn("No events created for dynamic simulation")
            else:
                event_handler = ShcEvents(self.app)

                # Remove any other events saved before
                event_handler.clear_all()
                
                logger.info("Creating your events.. please wait")
                for num, event in enumerate(events):
                    resp = event_handler.create(event)
                    if not resp:
                        raise Exception("Something failed when creating your event") 
                    else:
                        logger.info("Event created successfully") 
                        num += 1
                
            # Setting dynamic sim properties
            logger.warn(f"Preparing dynamic simulation {sim_type} with {num} event(s)")
            inc = self.app.GetFromStudyCase("ComInc")
            sim = self.app.GetFromStudyCase("ComSim")
            inc.p_resvar = self.app.GetFromStudyCase('TmpSimRel.ElmRes')
            inc.iopt_show = 1
            inc.iopt_sim = sim_type
            inc.tstart = start_time
            inc.dtgrd = step
            sim.tstop = end_time

            # Prepara results
            sim_res = ResDynSim(self.app)
            sim_res._prepare_results()

            # Init initial conditions
            is_inc_ok = bool(inc.Execute())
            # if is_inc_ok == False:
            #     logger.warn("Initial conditions might have issues")

            # Run RMS sim
            is_sim_ok = bool(sim.Execute())

            if is_sim_ok == True:
                logger.info(f"{sim_type.upper()} simulation - OK -")
            else:
                logger.warn(f"{sim_type.upper()} simulation - might have problems -")

            if events:
                # Events will be deleted
                if del_events:
                    for event in events:
                        event_handler.delete(event)
                # Events will be paused
                else:
                    for event in events:
                        event_handler.pause(event) 
        else:
            logger.warn("Dynamic simulation cannot executed .. activate a StudyCase")

        cond = sim_res._get_results()

        return sim_res if cond else None

    def run_short_circuit(self):
        raise Exception("Not implemented yet")


import os
import sys
import pandas as pd
import numpy as np
from ..utils import logger


class ResLdf:

    def __init__(self, app, ldf_cond):
        
        self.app = app

        if ldf_cond:
            pass
        else:
            raise Exception("Load flow failed")

    def attr_valid(self, attr, typ, dec=2):
        try:
            val = typ.GetAttribute(attr)
            val = round(val, dec)
        except:
            val = np.nan
        return val

    @property
    def line(self):

        line_dict = {}
        lines = self.app.GetCalcRelevantObjects('*.ElmLne')
        for line in lines:

            try:
                node_bus1 = line.bus1.cBusBar.loc_name
            except:
                node_bus1 = None
            try:
                node_bus2 = line.bus2.cBusBar.loc_name
            except:
                node_bus2 = None
            
            params = {
                "node_bus1": node_bus1,
                "node_bus2": node_bus2,
                "loading": self.attr_valid(attr="c:loading", typ=line),
                "losses": self.attr_valid(attr="c:Losses", typ=line),
                "Vbus1": self.attr_valid(attr="m:U1:bus1", typ=line),
                "Vbus2": self.attr_valid(attr="m:U1:bus2", typ=line),
                "angbus1": self.attr_valid(attr="m:phiu1:bus1", typ=line),
                "angbus2": self.attr_valid(attr="m:phiu1:bus2", typ=line),
                "Ibus1": self.attr_valid(attr="m:I1:bus1", typ=line),
                "Ibus2": self.attr_valid(attr="m:I1:bus2", typ=line),
                "Psumbus1": self.attr_valid(attr="m:Psum:bus1", typ=line),
                "Psumbus2": self.attr_valid(attr="m:Psum:bus2", typ=line),
                "Qsumbus1": self.attr_valid(attr="m:Qsum:bus1", typ=line),
                "Qsumbus2": self.attr_valid(attr="m:Qsum:bus2", typ=line),
            } 
            line_dict[line.loc_name] = params

        return pd.DataFrame.from_dict(line_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def line_pu(self):

        line_dict = {}
        lines = self.app.GetCalcRelevantObjects('*.ElmLne')
        for line in lines:

            try:
                node_bus1 = line.bus1.cBusBar.loc_name
            except:
                node_bus1 = None
            try:
                node_bus2 = line.bus2.cBusBar.loc_name
            except:
                node_bus2 = None
            
            params = {
                "node_bus1": node_bus1,
                "node_bus2": node_bus2,
                "loading": self.attr_valid(attr="c:loading", typ=line),
                "vbus1": self.attr_valid(attr="m:u1:bus1", typ=line),
                "vbus2": self.attr_valid(attr="m:u1:bus2", typ=line),
                "ibus1": self.attr_valid(attr="m:i1:bus1", typ=line),
                "ibus2": self.attr_valid(attr="m:i1:bus2", typ=line),
            } 
            line_dict[line.loc_name] = params

        return pd.DataFrame.from_dict(line_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def gen(self):
        
        gen_dict = {}
        gens = self.app.GetCalcRelevantObjects('*.ElmSym')
        for gen in gens:

            try:
                node = gen.bus1.cBusBar.loc_name
            except:
                node = None

            typ = gen
            params = {
                "node": node,
                "loading": self.attr_valid(attr="c:loading", typ=typ),
                "S": self.attr_valid(attr="e:sgini", typ=typ),
                "P": self.attr_valid(attr="e:pgini", typ=typ),
                "Q": self.attr_valid(attr="e:qgini", typ=typ),
                "pfactor": self.attr_valid(attr="e:cosgini", typ=typ),
                "V": self.attr_valid(attr="m:U1:bus1", typ=typ),
                "I": self.attr_valid(attr="m:I:bus1", typ=typ),
            }
            gen_dict[gen.loc_name] = params

        return pd.DataFrame.from_dict(gen_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def gen_pu(self):
        
        gen_dict = {}
        gens = self.app.GetCalcRelevantObjects('*.ElmSym')
        for gen in gens:

            try:
                node = gen.bus1.cBusBar.loc_name
            except:
                node = None

            typ = gen
            params = {
                "node": node,
                "v": self.attr_valid(attr="m:u1:bus1", typ=typ),
                "ang_v": self.attr_valid(attr="m:phiu1:bus1", typ=typ),
                "i": self.attr_valid(attr="m:i1:bus1", typ=typ),
                "ang_i": self.attr_valid(attr="m:phii1:bus1", typ=typ),
                "ang_vi": self.attr_valid(attr="m:phiui:bus1", typ=typ),
            }
            gen_dict[gen.loc_name] = params

        return pd.DataFrame.from_dict(gen_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def bus(self):

        bus_dict = {}
        buses = self.app.GetCalcRelevantObjects('*.ElmTerm')
        for bus in buses:

            try:
                node = bus.bus1.cBusBar.loc_name
            except:
                node = None

            typ = bus
            params = {
                "node": node,
                "V": self.attr_valid(attr="m:UI", typ=typ),
                "ang": self.attr_valid(attr="m:phiu", typ=typ),
            }
            bus_dict[bus.loc_name] = params

        return pd.DataFrame.from_dict(bus_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def bus_pu(self):

        bus_dict = {}
        buses = self.app.GetCalcRelevantObjects('*.ElmTerm')
        for bus in buses:

            try:
                node = bus.bus1.cBusBar.loc_name
            except:
                node = None

            typ = bus
            params = {
                "node": node,
                "v": self.attr_valid(attr="m:u", typ=typ),
                "ang": self.attr_valid(attr="m:phiu", typ=typ),
            }
            bus_dict[bus.loc_name] = params

        return pd.DataFrame.from_dict(bus_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})
    
    @property
    def trafo3w(self):
        
        trafo_dict = {}
        trafos = self.app.GetCalcRelevantObjects('*.ElmTr3')
        for trafo in trafos:

            try:
                node_hv = trafo.bushv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_mv = trafo.busmv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_lv = trafo.buslv.cBusBar.loc_name
            except:
                node_hv = np.nan

            typ = trafo
            params = {
                "node_hv": node_hv,
                "node_mv": node_mv,
                # "node_lv": node_lv,
                "loading_h": self.attr_valid(attr="c:loading_h", typ=typ),
                "loading_m": self.attr_valid(attr="c:loading_m", typ=typ),
                # "loading_l": self.attr_valid(attr="c:loading_l", typ=typ),
                "tot_ploss": self.attr_valid(attr="c:Ploss", typ=typ),
                "tot_qloss": self.attr_valid(attr="c:Qloss", typ=typ),
                "Vhv": self.attr_valid(attr="m:U1l:bushv", typ=typ),
                "Vmv": self.attr_valid(attr="m:U1l:busmv", typ=typ),
                # "Vlv": self.attr_valid(attr="m:U1l:buslv", typ=typ),
                "ang_vhv": self.attr_valid(attr="m:phiu1:bushv", typ=typ),
                "ang_vmv": self.attr_valid(attr="m:phiu1:busmv", typ=typ),
                # "ang_vlv": self.attr_valid(attr="m:phiu1:buslv", typ=typ),
                "Ihv": self.attr_valid(attr="m:I:bushv", typ=typ),
                "Imv": self.attr_valid(attr="m:I:busmv", typ=typ),
                # "Ilv": self.attr_valid(attr="m:I:buslv", typ=typ),
                "ang_ihv": self.attr_valid(attr="m:phii:bushv", typ=typ),
                "ang_imv": self.attr_valid(attr="m:phii:busmv", typ=typ),
                # "ang_ilv": self.attr_valid(attr="m:phii:buslv", typ=typ),
                "Phv": self.attr_valid(attr="m:Psum:bushv", typ=typ),
                "Pmv": self.attr_valid(attr="m:Psum:busmv", typ=typ),
                # "Plv": self.attr_valid(attr="m:Psum:buslv", typ=typ),
                "Qhv": self.attr_valid(attr="m:Qsum:bushv", typ=typ),
                "Qmv": self.attr_valid(attr="m:Qsum:busmv", typ=typ),
                # "Qlv": self.attr_valid(attr="m:Qsum:buslv", typ=typ),

            }
            trafo_dict[trafo.loc_name] = params

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def trafo3w_pu(self):
        
        trafo_dict = {}
        trafos = self.app.GetCalcRelevantObjects('*.ElmTr3')
        for trafo in trafos:

            try:
                node_hv = trafo.bushv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_mv = trafo.busmv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_lv = trafo.buslv.cBusBar.loc_name
            except:
                node_hv = np.nan

            typ = trafo
            params = {
                "node_hv": node_hv,
                "node_mv": node_mv,
                # "node_lv": node_lv,
                "vhv": self.attr_valid(attr="m:u1:bushv", typ=typ),
                "vmv": self.attr_valid(attr="m:u1:busmv", typ=typ),
                # "vlv": self.attr_valid(attr="m:u1:buslv", typ=typ),
                "ang_vhv": self.attr_valid(attr="m:phiu1:bushv", typ=typ),
                "ang_vmv": self.attr_valid(attr="m:phiu1:busmv", typ=typ),
                # "ang_vlv": self.attr_valid(attr="m:phiu1:buslv", typ=typ),
                "ihv": self.attr_valid(attr="m:i1:bushv", typ=typ),
                "imv": self.attr_valid(attr="m:i1:busmv", typ=typ),
                # "ilv": self.attr_valid(attr="m:i1:buslv", typ=typ),
                "ang_ihv": self.attr_valid(attr="m:phii:bushv", typ=typ),
                "ang_imv": self.attr_valid(attr="m:phii:busmv", typ=typ),
                # "ang_ilv": self.attr_valid(attr="m:phii:buslv", typ=typ),
            }
            trafo_dict[trafo.loc_name] = params

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})
    
    @property
    def trafo2w(self):
        
        trafo_dict = {}
        trafos = self.app.GetCalcRelevantObjects('*.ElmTr2')
        for trafo in trafos:

            try:
                node_hv = trafo.bushv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_lv = trafo.buslv.cBusBar.loc_name
            except:
                node_hv = np.nan

            typ = trafo
            params = {
                "node_hv": node_hv,
                "node_lv": node_lv,
                "loading_h": self.attr_valid(attr="c:loading_h", typ=typ),
                "loading_l": self.attr_valid(attr="c:loading_l", typ=typ),
                "tot_ploss": self.attr_valid(attr="c:Ploss", typ=typ),
                "tot_qloss": self.attr_valid(attr="c:Qloss", typ=typ),
                "Vhv": self.attr_valid(attr="m:U1l:bushv", typ=typ),
                "Vlv": self.attr_valid(attr="m:U1l:bushv", typ=typ),
                "ang_vhv": self.attr_valid(attr="m:phiu1:bushv", typ=typ),
                "ang_vlv": self.attr_valid(attr="m:phiu1:buslv", typ=typ),
                "Ihv": self.attr_valid(attr="m:I:bushv", typ=typ),
                "Ilv": self.attr_valid(attr="m:I:buslv", typ=typ),
                "ang_ihv": self.attr_valid(attr="m:phii:bushv", typ=typ),
                "ang_ilv": self.attr_valid(attr="m:phii:buslv", typ=typ),
                "Phv": self.attr_valid(attr="m:Psum:bushv", typ=typ),
                "Plv": self.attr_valid(attr="m:Psum:buslv", typ=typ),
                "Qhv": self.attr_valid(attr="m:Qsum:bushv", typ=typ),
                "Qlv": self.attr_valid(attr="m:Qsum:buslv", typ=typ),
            }
            trafo_dict[trafo.loc_name] = params

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def trafo2w_pu(self):
        
        trafo_dict = {}
        trafos = self.app.GetCalcRelevantObjects('*.ElmTr2')
        for trafo in trafos:

            try:
                node_hv = trafo.bushv.cBusBar.loc_name
            except:
                node_hv = np.nan
            try:
                node_lv = trafo.buslv.cBusBar.loc_name
            except:
                node_hv = np.nan

            typ = trafo
            params = {
                "node_hv": node_hv,
                "node_lv": node_lv,
                "vhv": self.attr_valid(attr="m:u1:bushv", typ=typ),
                "vlv": self.attr_valid(attr="m:u1:buslv", typ=typ),
                "ang_vhv": self.attr_valid(attr="m:phiu1:bushv", typ=typ),
                "ang_vlv": self.attr_valid(attr="m:phiu1:buslv", typ=typ),
                "ihv": self.attr_valid(attr="m:i1:bushv", typ=typ),
                "ilv": self.attr_valid(attr="m:i1:buslv", typ=typ),
                "ang_ihv": self.attr_valid(attr="m:phii:bushv", typ=typ),
                "ang_ilv": self.attr_valid(attr="m:phii:buslv", typ=typ),
            }
            trafo_dict[trafo.loc_name] = params

        return pd.DataFrame.from_dict(trafo_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})

    @property
    def shunt(self):
        
        shunt_dict = {}
        shunts = self.app.GetCalcRelevantObjects('*.ElmShnt')
        for shunt in shunts:

            try:
                node = shunt.bushv.cBusBar.loc_name
            except:
                node = np.nan

            typ = shunt
            params = {
                "node": node,
                "V": self.attr_valid(attr="m:U1l:bus1", typ=typ),
                "Q": self.attr_valid(attr="m:Q:bus1", typ=typ),
                "v": self.attr_valid(attr="m:u1:bus1", typ=typ),
                "ang_vi": self.attr_valid(attr="m:phiui:bus1", typ=typ),
            }
            shunt_dict[shunt.loc_name] = params

        return pd.DataFrame.from_dict(shunt_dict, orient="index") \
                           .reset_index() \
                           .rename(columns={"index": "name"})


from setuptools import setup, find_packages

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()
    
 
classifiers = [
  'Development Status :: 1 - Planning',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
  'Programming Language :: Python :: 3',
]
 
setup(
  name='yapowf',
  version='0.0.4',
  description='A new Python Power Factory API',
  long_description=long_description,
  url='',  
  author='Camilo Romero',
  author_email='camilo.romers@gmail.com',
  license='GNU GPLv3', 
  classifiers=classifiers,
  keywords=['power factory api', 'digsilent python'], 
  packages=find_packages(),
  install_requires=['pandas'] 
)